#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
from typing import Any, Callable, Dict, List, Optional, Union, cast

from boto3 import Session

from seedfarmer.services._service_utils import boto3_client, get_region, get_sts_identity_info


def get_secrets_manager_value(
    name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> Dict[str, Any]:
    account_id, _, partition = get_sts_identity_info(session=session)
    client = boto3_client(service_name="secretsmanager", session=session)
    secret_arn = f"arn:{partition}:secretsmanager:{get_region(session=session)}:{account_id}:secret:{name}"
    json_str: str = client.get_secret_value(SecretId=secret_arn)["SecretString"]
    return cast(Dict[str, Any], json.loads(json_str))


def list_secret_version_ids(
    name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> Optional[List[Any]]:
    account_id, _, partition = get_sts_identity_info(session=session)
    client = boto3_client(service_name="secretsmanager", session=session)
    secret_arn = f"arn:{partition}:secretsmanager:{get_region(session=session)}:{account_id}:secret:{name}"
    resp = client.list_secret_version_ids(SecretId=secret_arn)
    return client.list_secret_version_ids(SecretId=secret_arn)["Versions"] if resp else None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from enum import Enum
from typing import Any, Dict, Optional

from pydantic import PrivateAttr

from seedfarmer.models._base import CamelModel
from seedfarmer.utils import generate_codebuild_url


class StatusType(Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


class CodeBuildMetadata(CamelModel):
    _build_url: str = PrivateAttr()

    aws_account_id: Optional[str] = None
    aws_region: Optional[str] = None
    aws_partition: Optional[str] = None
    codebuild_build_id: Optional[str] = None
    codebuild_log_path: Optional[str] = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._build_url = generate_codebuild_url(
            account_id=self.aws_account_id,  # type: ignore
            region=self.aws_region,  # type: ignore
            codebuild_id=self.codebuild_build_id,  # type: ignore
            partition=self.aws_partition,
        )

    @property
    def build_url(self) -> str:
        return self._build_url


class ModuleDeploymentResponse(CamelModel):
    deployment: str
    group: Optional[str] = None
    module: str
    status: str
    codebuild_metadata: Optional[CodeBuildMetadata] = None
    codebuild_output: Optional[Dict[Any, Any]] = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        if data.get("status"):
            status_in = str(data["status"]).upper()
            self.status = status_in if status_in in StatusType.__members__.keys() else StatusType.SUCCESS.value
        else:
            self.build_type = StatusType.SUCCESS.value



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
from typing import Any, Dict, List, cast

import seedfarmer.errors
import seedfarmer.services._ssm as ssm
from seedfarmer.models import ValueFromRef
from seedfarmer.models.manifests import NetworkMapping
from seedfarmer.services.session_manager import SessionManager

_logger: logging.Logger = logging.getLogger(__name__)


def load_network_values(
    network: NetworkMapping, parameters_regional: Dict[str, Any], account_id: str, region: str
) -> NetworkMapping:
    _logger.debug("Evaluating network for  %s and %s", account_id, region)

    if network.vpc_id is None and network.private_subnet_ids is None and network.security_group_ids is None:
        _logger.debug("None of the parameters are populated, ignoring networking")
        return network

    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)

    if isinstance(network.vpc_id, ValueFromRef):
        if network.vpc_id.value_from and network.vpc_id.value_from.parameter_value:
            network.vpc_id = cast(
                str,
                parameters_regional.get(str(network.vpc_id.value_from.parameter_value)),
            )
        elif network.vpc_id.value_from and network.vpc_id.value_from.parameter_store:
            network.vpc_id = cast(
                str, ssm.get_parameter(name=str(network.vpc_id.value_from.parameter_store), session=session)
            )
        elif network.vpc_id.value_from and network.vpc_id.value_from.env_variable:
            network.vpc_id = cast(str, os.getenv(network.vpc_id.value_from.env_variable, ""))

    if isinstance(network.private_subnet_ids, ValueFromRef):
        if network.private_subnet_ids.value_from and network.private_subnet_ids.value_from.parameter_value:
            network.private_subnet_ids = cast(
                List[str],
                parameters_regional.get(str(network.private_subnet_ids.value_from.parameter_value)),
            )
        elif network.private_subnet_ids.value_from and network.private_subnet_ids.value_from.parameter_store:
            network.private_subnet_ids = cast(
                List[str],
                ssm.get_parameter(name=str(network.private_subnet_ids.value_from.parameter_store), session=session),
            )
        elif network.private_subnet_ids.value_from and network.private_subnet_ids.value_from.env_variable:
            network.private_subnet_ids = cast(
                List[str],
                json.loads(os.getenv(network.private_subnet_ids.value_from.env_variable, "")),
            )
    if isinstance(network.security_group_ids, ValueFromRef):
        if network.security_group_ids.value_from and network.security_group_ids.value_from.parameter_value:
            network.security_group_ids = cast(
                List[str],
                parameters_regional.get(str(network.security_group_ids.value_from.parameter_value)),
            )
        elif network.security_group_ids.value_from and network.security_group_ids.value_from.parameter_store:
            network.security_group_ids = cast(
                List[str],
                ssm.get_parameter(name=str(network.security_group_ids.value_from.parameter_store), session=session),
            )
        elif network.security_group_ids.value_from and network.security_group_ids.value_from.env_variable:
            network.security_group_ids = cast(
                List[str],
                json.loads(os.getenv(network.security_group_ids.value_from.env_variable, "")),
            )

        if len(network.security_group_ids) > 5:  # type: ignore
            raise seedfarmer.errors.InvalidConfigurationError("Cannot have more than 5 Security Groups in a Network")

    _logger.debug("Returning network for %s - %s : %s", account_id, region, network)
    return network



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import random
import string
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional, Union

from boto3 import Session

import seedfarmer.services._cloudwatch as cloudwatch
import seedfarmer.services._codebuild as codebuild
import seedfarmer.services._s3 as s3

_logger: logging.Logger = logging.getLogger(__name__)


def _print_codebuild_logs(
    events: List[cloudwatch.CloudWatchEvent],
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
) -> None:
    for event in events:
        msg = event.message[:-1] if event.message.endswith("\n") else event.message
        _logger.debug("[CODEBUILD] %s", msg)
        if codebuild_log_callback:
            codebuild_log_callback(msg)


def _wait_execution(
    build_id: str,
    stream_name_prefix: str,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
) -> Optional[codebuild.BuildInfo]:
    start_time: Optional[datetime] = None
    stream_name: Optional[str] = None
    status: Optional[codebuild.BuildInfo] = None
    for status in codebuild.wait(build_id=build_id, session=session):
        if status.logs.enabled and status.logs.group_name:
            if stream_name is None:
                stream_name = cloudwatch.get_stream_name_by_prefix(
                    group_name=status.logs.group_name, prefix=f"{stream_name_prefix}/", session=session
                )
            if stream_name is not None:
                events = cloudwatch.get_log_events(
                    group_name=status.logs.group_name, stream_name=stream_name, start_time=start_time, session=session
                )
                _print_codebuild_logs(events=events.events, codebuild_log_callback=codebuild_log_callback)
                if events.last_timestamp is not None:
                    start_time = events.last_timestamp + timedelta(milliseconds=1)
    return status


def _execute_codebuild(
    stack_outputs: Dict[str, str],
    bundle_location: str,
    execution_id: str,
    buildspec: Dict[str, Any],
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    yaml_dumper: Optional[Any] = None,  # Accepts ruamel.yaml.YAML instance or PyYAML dump function
) -> Optional[codebuild.BuildInfo]:
    _logger.debug("bundle_location: %s", bundle_location)
    stream_name_prefix = f"codeseeder-{execution_id}"  # (LEGACY)
    _logger.debug("stream_name_prefix: %s", stream_name_prefix)
    build_id = codebuild.start(
        project_name=stack_outputs["CodeBuildProject"],
        stream_name=stream_name_prefix,
        bundle_location=bundle_location,
        buildspec=buildspec,
        timeout=timeout,
        overrides=overrides,
        session=session,
        yaml_dumper=yaml_dumper,
    )
    return _wait_execution(
        build_id=build_id,
        stream_name_prefix=stream_name_prefix,
        codebuild_log_callback=codebuild_log_callback,
        session=session,
    )


def run(
    stack_outputs: Dict[str, str],
    bundle_path: str,
    buildspec: Dict[str, Any],
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    bundle_id: Optional[str] = None,
    prebuilt_bundle: Optional[str] = None,
    yaml_dumper: Optional[Any] = None,  # Accepts ruamel.yaml.YAML instance or PyYAML dump function
) -> Optional[codebuild.BuildInfo]:
    execution_id = "".join(random.choice(string.ascii_lowercase) for i in range(8))

    if prebuilt_bundle:
        prebuilt_bundle = prebuilt_bundle[5:] if prebuilt_bundle.startswith("s3://") else prebuilt_bundle
        o = prebuilt_bundle.split("/", 1)
        loc = f"{o[0]}/{o[1]}"
    else:
        key: str = (
            f"seedfarmer/{bundle_id}/{execution_id}/bundle.zip"
            if bundle_id
            else f"seedfarmer/{execution_id}/bundle.zip"
        )
        bucket = stack_outputs["Bucket"]
        s3.delete_objects(bucket=bucket, keys=[key], session=session)
        s3.upload_file(src=bundle_path, bucket=bucket, key=key, session=session)
        loc = f"{bucket}/{key}"

    build_info = _execute_codebuild(
        stack_outputs=stack_outputs,
        bundle_location=loc,
        execution_id=execution_id,
        buildspec=buildspec,
        codebuild_log_callback=codebuild_log_callback,
        timeout=timeout,
        overrides=overrides,
        session=session,
        yaml_dumper=yaml_dumper,
    )
    if not prebuilt_bundle:
        s3.delete_objects(bucket=bucket, keys=[key], session=session)
    return build_info



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Any, Dict, List, Optional

from pydantic import PrivateAttr, model_validator
from pydantic.json_schema import SkipJsonSchema

from seedfarmer.errors import InvalidManifestError
from seedfarmer.models._base import CamelModel, ValueFromRef
from seedfarmer.models._deploy_spec import DeploySpec
from seedfarmer.utils import upper_snake_case


class ModuleParameter(ValueFromRef):
    _upper_snake_case: str = PrivateAttr()

    name: str
    value: Optional[Any] = None
    version: Optional[Any] = None
    disableEnvVarResolution: Optional[bool] = None
    resolved_value: Optional[Any] = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._upper_snake_case = upper_snake_case(self.name)

    @property
    def upper_snake_case(self) -> str:
        return self._upper_snake_case

    @model_validator(mode="after")
    def check_value_or_value_from(self) -> "ModuleParameter":
        value = self.value
        value_from = self.value_from

        if value is None and value_from is None:
            raise InvalidManifestError(f"value or value_from must be provided for parameter {self.name}")
        if value is not None and value_from is not None:
            raise InvalidManifestError(f"value and value_from cannot be provided for parameter {self.name}")

        return self


class DataFile(CamelModel):
    file_path: str
    commit_hash: SkipJsonSchema[Optional[str]] = None
    _local_file_path: Optional[str] = PrivateAttr(default=None)
    _bundle_path: Optional[str] = PrivateAttr(default=None)

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._local_file_path = "."
        self._bundle_path = self.file_path

    def set_local_file_path(self, _local_file_path: str) -> None:
        self._local_file_path = _local_file_path

    def get_local_file_path(self) -> Optional[str]:
        return self._local_file_path

    def set_bundle_path(self, bundle_path: str) -> None:
        self._bundle_path = bundle_path

    def get_bundle_path(self) -> Optional[str]:
        return self._bundle_path


class ModuleManifest(CamelModel):
    """
    ModuleManifest
    This is a module in a group of the deployment and consists of a name,
    the path of the module manifest, any ModuleParameters for deployment, and
    the DeploySpec.
    """

    name: str
    path: str
    bundle_md5: Optional[str] = None
    manifest_md5: Optional[str] = None
    deployspec_md5: Optional[str] = None
    parameters: List[ModuleParameter] = []
    deploy_spec: Optional[DeploySpec] = None
    target_account: Optional[str] = None
    target_region: Optional[str] = None
    codebuild_image: Optional[str] = None
    runtime_overrides: Optional[Dict[str, str]] = None
    data_files: Optional[List[DataFile]] = None
    commit_hash: SkipJsonSchema[Optional[str]] = None
    npm_mirror: Optional[str] = None
    npm_mirror_secret: Optional[str] = None
    pypi_mirror: Optional[str] = None
    pypi_mirror_secret: Optional[str] = None
    _target_account_id: Optional[str] = PrivateAttr(default=None)
    _local_path: Optional[str] = PrivateAttr(default=None)

    def __init__(self, **kwargs: Any) -> None:
        from seedfarmer.utils import batch_replace_env

        kwargs = batch_replace_env(payload=kwargs)
        super().__init__(**kwargs)
        self._local_path = self.path

    def set_target_account_id(self, account_id: str) -> None:
        self._target_account_id = account_id

    def get_target_account_id(self) -> Optional[str]:
        return self._target_account_id

    def set_local_path(self, local_path: str) -> None:
        self._local_path = local_path

    def get_local_path(self) -> Optional[str]:
        return self._local_path



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import logging
from typing import Optional

from boto3 import Session

import seedfarmer.services._s3 as s3
from seedfarmer import config

_logger: logging.Logger = logging.getLogger(__name__)


BUNDLE_PREFIX = "bundle"


class BundleS3Support:
    seedkit_bucket: Optional[str] = None  # NEEDED This is for the seedkit along with seedfarmer_bucket
    seedkit_key: Optional[str] = None  # NEEDED This is for the seedkit along wih seedfarmer_bucket
    seedfarmer_bucket: Optional[str] = None
    seedfarmer_key: Optional[str] = None

    def __init__(
        self, deployment: str, group: str, module: str, bucket: str, bundle_src_path: Optional[str] = None
    ) -> None:
        self.project = config.PROJECT
        self.ops_root_path = config.OPS_ROOT
        self.seedfarmer_bucket = bucket
        self.seedfarmer_key = f"{BUNDLE_PREFIX}/{deployment}/{group}/{module}/bundle.zip"

        if bundle_src_path is not None:
            o = bundle_src_path.split("/", 1)
            self.seedkit_bucket = o[0]
            self.seedkit_key = o[1]


def copy_bundle_to_sf(
    deployment: str, group: str, module: str, bucket: str, bundle_src_path: str, session: Optional[Session] = None
) -> None:
    bundle = BundleS3Support(
        deployment=deployment, group=group, module=module, bucket=bucket, bundle_src_path=bundle_src_path
    )
    try:
        s3.copy_s3_object(
            src_bucket=str(bundle.seedkit_bucket),
            src_key=str(bundle.seedkit_key),
            dest_bucket=str(bundle.seedfarmer_bucket),
            dest_key=str(bundle.seedfarmer_key),
            session=session,
        )
    except Exception as e:
        _logger.info("Cannot copy the bundle to S3 - %s", e)


def get_bundle_sf_path(
    deployment: str,
    group: str,
    module: str,
    bucket: str,
) -> str:
    bundle = BundleS3Support(deployment=deployment, group=group, module=module, bucket=bucket)
    return f"s3://{bundle.seedfarmer_bucket}/{bundle.seedfarmer_key}"


def delete_bundle_from_sf(
    deployment: str,
    group: str,
    module: str,
    bucket: str,
    session: Optional[Session] = None,
) -> None:
    bundle = BundleS3Support(deployment=deployment, group=group, module=module, bucket=bucket)
    try:
        s3.delete_objects(bucket=str(bundle.seedfarmer_bucket), keys=[str(bundle.seedfarmer_key)], session=session)
    except Exception as e:
        _logger.info("Cannot delete the bundle from S3 - %s", e)


def check_bundle_exists_in_sf(
    deployment: str,
    group: str,
    module: str,
    bucket: str,
    session: Optional[Session] = None,
) -> bool:
    bundle = BundleS3Support(deployment=deployment, group=group, module=module, bucket=bucket)
    try:
        return bool(
            s3.object_exists(bucket=str(bundle.seedfarmer_bucket), key=str(bundle.seedfarmer_key), session=session)
        )
    except Exception as e:
        _logger.info("Cannot check if the bundle exists in S3 - %s", e)
        return False



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from seedfarmer.models.manifests._deployment_manifest import (
    DeploymentManifest,
    ModulesManifest,
    NameGenerator,
    NetworkMapping,
    RegionMapping,
    TargetAccountMapping,
)
from seedfarmer.models.manifests._module_manifest import DataFile, ModuleManifest, ModuleParameter

__all__ = [
    "DeploymentManifest",
    "ModulesManifest",
    "NameGenerator",
    "NetworkMapping",
    "RegionMapping",
    "TargetAccountMapping",
    "ModuleManifest",
    "ModuleParameter",
    "DataFile",
]



#    Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from enum import Enum
from typing import Dict, Optional


class CuratedBuildImages:
    class ImageEnums(Enum):
        UBUNTU_STANDARD_6 = "aws/codebuild/standard:6.0"
        UBUNTU_STANDARD_7 = "aws/codebuild/standard:7.0"
        AL2_X86_64_STANDARD_5_0 = "aws/codebuild/amazonlinux2-x86_64-standard:5.0"
        AL2_AARCH64_STANDARD_3_0 = "aws/codebuild/amazonlinux2-aarch64-standard:3.0"

    class ImageRuntimes(Enum):
        UBUNTU_STANDARD_6 = {"nodejs": "16", "python": "3.10", "java": "corretto17"}
        UBUNTU_STANDARD_7 = {"nodejs": "18", "python": "3.11", "java": "corretto21"}
        AL2_X86_64_STANDARD_5_0 = {"nodejs": "20", "python": "3.12", "java": "corretto21"}
        AL2_AARCH64_STANDARD_3_0 = {"nodejs": "20", "python": "3.12", "java": "corretto21"}


def get_runtimes(
    codebuild_image: Optional[str], runtime_overrides: Optional[Dict[str, str]] = None
) -> Optional[Dict[str, str]]:
    runtime_overrides = runtime_overrides if runtime_overrides else {}
    image_vals = [cbi.value for cbi in CuratedBuildImages.ImageEnums]
    if codebuild_image in image_vals:
        cir_d = {cir.name: cir.value for cir in CuratedBuildImages.ImageRuntimes}
        k = CuratedBuildImages.ImageEnums(codebuild_image).name
        return {**cir_d[k], **runtime_overrides}
    return runtime_overrides



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import json
import logging
import os
import re
import sys
from typing import Any, Dict, List, Optional, Tuple, cast

import yaml
from boto3 import Session
from botocore.exceptions import WaiterError

import seedfarmer.errors
import seedfarmer.services._cfn as cfn
import seedfarmer.services._iam as iam
from seedfarmer import CLI_ROOT, __version__
from seedfarmer.services import create_new_session, get_region, get_sts_identity_info
from seedfarmer.utils import get_deployment_role_name, get_toolchain_role_arn, get_toolchain_role_name, valid_qualifier

_logger: logging.Logger = logging.getLogger(__name__)


def get_template(template_name: str) -> Dict[str, Any]:
    with open((os.path.join(CLI_ROOT, f"resources/{template_name}.template")), "r") as f:
        role = yaml.safe_load(f)

    if not isinstance(role, dict):
        raise TypeError(f"Expected dictionary from YAML file, got {type(role).__name__}")
    return role


def write_template(template: Dict[Any, Any]) -> None:
    yaml.dump(template, sys.stdout)
    print("")


def bootstrap_toolchain_account(
    project_name: str,
    principal_arns: List[str],
    permissions_boundary_arn: Optional[str] = None,
    policy_arns: Optional[List[str]] = None,
    qualifier: Optional[str] = None,
    role_prefix: Optional[str] = None,
    policy_prefix: Optional[str] = None,
    profile: Optional[str] = None,
    region_name: Optional[str] = None,
    synthesize: bool = False,
    as_target: bool = False,
) -> Optional[Dict[Any, Any]]:
    if qualifier and not valid_qualifier(qualifier):
        raise seedfarmer.errors.InvalidConfigurationError("The Qualifier must be alphanumeric and 6 characters or less")

    for arn in principal_arns:
        if not re.match(r"arn:aws.*:(sts|iam)::(\d{12}|\*):.*$", arn):
            raise seedfarmer.errors.InvalidConfigurationError(f"Trusted principal: {arn} is not a valid principal arn")

    role_stack_name = get_toolchain_role_name(project_name=project_name, qualifier=cast(str, qualifier))
    template = get_template("toolchain_role")
    _logger.debug((json.dumps(template, indent=4)))

    # Prepare CloudFormation parameters
    role_prefix_value = role_prefix if role_prefix else "/"
    parameters: Dict[str, str] = {
        "ProjectName": project_name,
        "RoleName": role_stack_name,
        "RolePrefix": role_prefix_value,
        "SeedFarmerVersion": __version__,
        "PrincipalArns": ",".join(principal_arns),
    }

    # Only add PermissionsBoundaryArn if it's not None
    if permissions_boundary_arn:
        parameters["PermissionsBoundaryArn"] = permissions_boundary_arn

    _logger.debug((json.dumps(parameters, indent=4)))

    if not synthesize:
        session = create_new_session(profile=profile, region_name=region_name)
        session_account_id, _, _ = get_sts_identity_info(session=session)
        apply_deploy_logic(
            template=template,
            role_name=role_stack_name,
            stack_name=role_stack_name,
            session=session,
            account_id=session_account_id,
            parameters=parameters,
        )
        if as_target:
            bootstrap_target_account(
                toolchain_account_id=session_account_id,
                project_name=project_name,
                qualifier=cast(str, qualifier),
                role_prefix=role_prefix,
                policy_prefix=policy_prefix,
                permissions_boundary_arn=permissions_boundary_arn,
                profile=profile,
                region_name=region_name,
                policy_arns=policy_arns,
                session=session,
            )
    else:
        write_template(template=template)
        if as_target:
            bootstrap_target_account(
                toolchain_account_id="123456789012",
                project_name=project_name,
                qualifier=cast(str, qualifier),
                permissions_boundary_arn=permissions_boundary_arn,
                policy_arns=policy_arns,
                profile=profile,
                region_name=region_name,
                session=None,
                synthesize=synthesize,
            )
    return template


def bootstrap_target_account(
    toolchain_account_id: str,
    project_name: str,
    permissions_boundary_arn: Optional[str] = None,
    qualifier: Optional[str] = None,
    role_prefix: Optional[str] = None,
    policy_prefix: Optional[str] = None,
    profile: Optional[str] = None,
    region_name: Optional[str] = None,
    session: Optional[Session] = None,
    policy_arns: Optional[List[str]] = None,
    synthesize: bool = False,
) -> Optional[Dict[Any, Any]]:
    if qualifier and not valid_qualifier(qualifier):
        raise seedfarmer.errors.InvalidConfigurationError("The Qualifier must be alphanumeric and 6 characters or less")

    if not session:
        session = create_new_session(profile=profile, region_name=region_name)
    session_account_id, _, partition = get_sts_identity_info(session=session)

    role_stack_name = get_deployment_role_name(project_name=project_name, qualifier=cast(str, qualifier))
    toolchain_role_arn = get_toolchain_role_arn(
        partition=partition,
        toolchain_account_id=toolchain_account_id,
        project_name=project_name,
        qualifier=cast(str, qualifier),
        role_prefix=role_prefix,
    )

    template = get_template("deployment_role")
    _logger.debug((json.dumps(template, indent=4)))

    # Prepare CloudFormation parameters
    role_prefix_value = role_prefix if role_prefix else "/"
    policy_prefix_value = policy_prefix if policy_prefix else "/"
    parameters: Dict[str, str] = {
        "RolePrefix": role_prefix_value,
        "PolicyPrefix": policy_prefix_value,
        "ProjectName": project_name,
        "RoleName": role_stack_name,
        "ToolchainRoleArn": toolchain_role_arn,
        "SeedFarmerVersion": __version__,
    }

    # Only add ManagedPolicyArn if it's not None
    if policy_arns:
        parameters["ManagedPolicyArns"] = ",".join(policy_arns)

    # Only add PermissionsBoundaryArn if it's not None
    if permissions_boundary_arn:
        parameters["PermissionsBoundaryArn"] = permissions_boundary_arn

    _logger.debug((json.dumps(parameters, indent=4)))

    if not synthesize:
        apply_deploy_logic(
            template=template,
            role_name=role_stack_name,
            stack_name=role_stack_name,
            session=session,
            account_id=session_account_id,
            parameters=parameters,
        )
    else:
        write_template(template=template)
    return template


def apply_deploy_logic(
    template: Dict[Any, Any],
    role_name: str,
    stack_name: str,
    session: Session,
    account_id: Optional[str] = None,
    parameters: Optional[Dict[str, str]] = None,
) -> None:
    role_exists, stack_exists = role_deploy_status(role_name=role_name, stack_name=stack_name, session=session)
    if not account_id:
        account_id, role_arn, partition = get_sts_identity_info(session=session)
    if not role_exists:
        _logger.info("Deploying role in account %s, region %s", account_id, get_region(session=session))
        deploy_template(template=template, stack_name=stack_name, session=session, parameters=parameters)
    else:
        if stack_exists[0]:
            _logger.info("Updating role in account %s, region %s", account_id, get_region(session=session))
            deploy_template(template=template, stack_name=stack_name, session=session, parameters=parameters)
        else:
            _logger.info(
                "The role %s exists in account %s as was not deployed in region %s, it will NOT be updated",
                role_name,
                account_id,
                get_region(session=session),
            )


def deploy_template(
    template: Dict[Any, Any], stack_name: str, session: Optional[Session], parameters: Optional[Dict[str, str]] = None
) -> None:
    loc = os.path.join(os.getcwd(), "templates")
    output = os.path.join(loc, f"{stack_name}.yaml")
    os.makedirs(loc, exist_ok=True)
    with open(output, "w") as outfile:
        yaml.dump(template, outfile)
    try:
        # Deploy with parameters
        cfn.deploy_template(stack_name=stack_name, filename=output, session=session, parameters=parameters)
        _logger.info(f"Role for Seed-Farmer deployed in stack {stack_name}")
    except WaiterError as we:
        _logger.error("Could not create the deployment role...make sure the toolchain role exists and check CFN Logs")
        raise we
    finally:
        os.remove(output)


def role_deploy_status(role_name: str, stack_name: str, session: Session) -> Tuple[Optional[Dict[str, Any]], Any]:
    return iam.get_role(role_name=role_name, session=session), cfn.does_stack_exist(
        stack_name=stack_name, session=session
    )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import concurrent.futures
import logging
import math
import random
import time
from itertools import repeat
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, cast

from boto3 import Session
from botocore.exceptions import ClientError

from seedfarmer.services._service_utils import boto3_client, boto3_resource

_logger: logging.Logger = logging.getLogger(__name__)

ChunkifyItemType = TypeVar("ChunkifyItemType")


def _chunkify(
    lst: List[ChunkifyItemType], num_chunks: int = 1, max_length: Optional[int] = None
) -> List[List[ChunkifyItemType]]:
    chunk_size: int = int(math.ceil(float(len(lst)) / float(num_chunks))) if max_length is None else max_length
    return [lst[i : i + chunk_size] for i in range(0, len(lst), chunk_size)]


def _delete_objects(
    bucket: str, chunk: List[Dict[str, str]], session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    client_s3 = boto3_client("s3", session=session)
    try:
        client_s3.delete_objects(Bucket=bucket, Delete={"Objects": chunk})  # type: ignore[typeddict-item]
    except client_s3.exceptions.ClientError as ex:
        if "SlowDown" in str(ex):
            time.sleep(random.randint(3, 10))
            client_s3.delete_objects(Bucket=bucket, Delete={"Objects": chunk})  # type: ignore[typeddict-item]


def list_keys(bucket: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> List[Dict[str, str]]:
    """List the keys/objects in an S3 Buket

    Parameters
    ----------
    bucket : str
        S3 Bucket name
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    List[Dict[str, str]]
        List of Keys and VersionIds
    """
    client_s3 = boto3_client("s3", session=session)
    paginator = client_s3.get_paginator("list_object_versions")
    response_iterator = paginator.paginate(Bucket=bucket, PaginationConfig={"PageSize": 1_000})
    keys: List[Dict[str, str]] = []
    for page in response_iterator:
        if "DeleteMarkers" in page:
            for delete_marker in page["DeleteMarkers"]:
                keys.append(
                    {
                        "Key": delete_marker["Key"],
                        "VersionId": delete_marker["VersionId"],
                    }
                )
        if "Versions" in page:
            for version in page["Versions"]:
                keys.append({"Key": version["Key"], "VersionId": version["VersionId"]})
    return keys


def delete_objects(
    bucket: str, keys: Optional[List[str]] = None, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    """Delete objects from an S3 Bucket

    Parameters
    ----------
    bucket : str
        S3 Bucket name
    keys : Optional[List[str]], optional
        List of keys to delete, all if None, by default None
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None
    """
    if keys is None:
        keys_pairs: List[Dict[str, str]] = list_keys(bucket=bucket, session=session)
    else:
        keys_pairs = [{"Key": k} for k in keys]
    if keys_pairs:
        chunks: List[List[Dict[str, str]]] = _chunkify(lst=keys_pairs, max_length=1_000)
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(chunks)) as executor:
            list(executor.map(_delete_objects, repeat(bucket), chunks, repeat(session)))


def delete_bucket(bucket: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> None:
    """Delete an S3 Bucket

    Parameters
    ----------
    bucket : str
        S3 Bucket Name
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Raises
    ------
    ex
        If error other that NoSuchBucket
    """
    client_s3 = boto3_client("s3", session=session)
    try:
        _logger.debug("Cleaning up bucket: %s", bucket)
        delete_objects(bucket=bucket, session=session)
        _logger.debug("Deleting bucket: %s", bucket)
        client_s3.delete_bucket(Bucket=bucket)
    except Exception as ex:
        if "NoSuchBucket" in str(ex):
            _logger.debug(f"Bucket ({bucket}) does not exist, skipping")
            return
        else:
            raise ex


def upload_file(
    src: str, bucket: str, key: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    """Upload file to S3 Bucket

    Parameters
    ----------
    src : str
        Local source file
    bucket : str
        S3 Bucket
    key : str
        Key name to upload to
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None
    """
    client_s3 = boto3_client("s3", session=session)
    client_s3.upload_file(Filename=src, Bucket=bucket, Key=key)


def list_s3_objects(
    bucket: str, prefix: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> Dict[str, Any]:
    """List S3 objects in a Bucket filtered to a prefix

    Parameters
    ----------
    bucket : str
        S3 Bucket name
    prefix : str
        Prefix filter
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    Dict[str, Any]
        List of objects
    """
    client_s3 = boto3_client("s3", session=session)
    response = client_s3.list_objects_v2(Bucket=bucket, Prefix=prefix)
    return cast(Dict[str, Any], response)


def delete_bucket_by_prefix(prefix: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> None:
    """Delete S3 Buckets whose name begins with a prefix

    Parameters
    ----------
    prefix : str
        Prefix to filter Buckets by
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None
    """
    client_s3 = boto3_client("s3", session=session)
    for bucket in client_s3.list_buckets()["Buckets"]:
        if bucket["Name"].startswith(prefix):
            delete_bucket(bucket=bucket["Name"], session=session)


def object_exists(bucket: str, key: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> bool:
    """Check for existence of an object in an S3 Bucket

    Parameters
    ----------
    bucket : str
        S3 Bucket name
    key : str
        Key to ckeck
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    bool
        Indicator of object existence
    """
    try:
        boto3_resource("s3", session=session).Object(bucket, key).load()  # type: ignore [attr-defined]
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        else:
            raise
    else:
        return True


def copy_s3_object(
    src_bucket: str,
    src_key: str,
    dest_bucket: str,
    dest_key: str,
    session: Optional[Union[Callable[[], Session], Session]] = None,
) -> None:
    """Copy an S3 Object to another bucket

    Parameters
    ----------
    src_bucket : str
        Source S3 Bucket
    src_key : str
        Source Key
    dest_bucket : str
        Destination S3 Bucket
    dest_key : str
        Destination Key
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None
    """
    client_s3 = boto3_client("s3", session=session)
    client_s3.copy_object(Bucket=dest_bucket, Key=dest_key, CopySource=f"{src_bucket}/{src_key}")


def is_bucket_empty(bucket: str, folder: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> bool:
    """
    Checks if there is any file in a specified folder within an S3 bucket.

    Parameters:
    - bucket: The name of the S3 bucket.
    - folder: The name of the folder inside the bucket.

    Returns:
    - False if the folder contains files, True otherwise.
    """
    if not folder.endswith("/"):
        folder += "/"
    client_s3 = boto3_client("s3", session=session)

    try:
        response = client_s3.list_objects_v2(Bucket=bucket, Prefix=folder)
        if "KeyCount" in response and response["KeyCount"] > 0:
            _logger.debug(f"Found {response['KeyCount']} objects under {folder}")
            return False
        else:
            return True
    except Exception as e:
        _logger.warn(f"Bucket ({bucket} - {e}) failed when looking for object count, have to assume there are objects")
        return False



#!/usr/bin/env python

import json
import logging
import os
import subprocess
from typing import Dict, cast

import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def get_secret() -> Dict[str, Dict[str, str]]:
    secret_name = os.environ.get("SEEDFARMER_DOCKER_SECRET", "NO_SECRET")
    # Backwards Compatibility
    if secret_name == "NO_SECRET":
        secret_name = os.environ.get("AWS_CODESEEDER_DOCKER_SECRET", "NO_SECRET")
    region_name = os.environ.get("AWS_DEFAULT_REGION")

    session = boto3.session.Session()
    client = session.client(service_name="secretsmanager", region_name=region_name)

    try:
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    except ClientError:
        logger.info("Secret with SecretId '%s' could not be retrieved from SecretsManager", secret_name)
        return {}
    else:
        return cast(Dict[str, Dict[str, str]], json.loads(get_secret_value_response.get("SecretString", "{}")))


if __name__ == "__main__":
    credentials = get_secret()
    for registry, creds in credentials.items():
        if not isinstance(creds, dict) or "username" not in creds or "password" not in creds:
            logger.error("Invalid credential structure for registry '%s'", registry)
            continue
        username = creds["username"]
        password = creds["password"]
        process = subprocess.Popen(
            ["docker", "login", "--username", username, "--password-stdin", registry], stdin=subprocess.PIPE, text=True
        )
        process.communicate(input=password)
        if process.returncode != 0:
            logger.error("Docker login failed for registry '%s'", registry)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from seedfarmer.commands._bootstrap_commands import bootstrap_target_account, bootstrap_toolchain_account
from seedfarmer.commands._deployment_commands import apply, destroy
from seedfarmer.commands._network_parameter_commands import load_network_values
from seedfarmer.commands._parameter_commands import generate_export_env_params, generate_export_raw_env_params
from seedfarmer.commands._project_policy_commands import get_default_project_policy
from seedfarmer.commands._stack_commands import (
    deploy_bucket_storage_stack,
    deploy_managed_policy_stack,
    deploy_module_stack,
    deploy_seedkit,
    destroy_bucket_storage_stack,
    destroy_managed_policy_stack,
    destroy_module_stack,
    destroy_seedkit,
    force_manage_policy_attach,
    get_module_stack_info,
)

__all__ = [
    "apply",
    "destroy",
    "deploy_managed_policy_stack",
    "destroy_managed_policy_stack",
    "deploy_module_stack",
    "destroy_module_stack",
    "deploy_seedkit",
    "destroy_seedkit",
    "get_module_stack_info",
    "generate_export_env_params",
    "generate_export_raw_env_params",
    "bootstrap_toolchain_account",
    "bootstrap_target_account",
    "get_default_project_policy",
    "load_network_values",
    "force_manage_policy_attach",
    "deploy_bucket_storage_stack",
    "destroy_bucket_storage_stack",
]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from seedfarmer.cli_groups._bootstrap_group import bootstrap
from seedfarmer.cli_groups._bundle_group import bundle
from seedfarmer.cli_groups._init_group import init
from seedfarmer.cli_groups._list_group import list
from seedfarmer.cli_groups._manage_metadata_group import metadata
from seedfarmer.cli_groups._project_group import projectpolicy
from seedfarmer.cli_groups._remove_group import remove
from seedfarmer.cli_groups._seedkit_group import seedkit
from seedfarmer.cli_groups._store_group import store
from seedfarmer.cli_groups._taint_group import taint

__all__ = ["bootstrap", "bundle", "init", "list", "metadata", "projectpolicy", "remove", "seedkit", "store", "taint"]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
from typing import Any, Callable, Dict, List, Optional, Union, cast

from boto3 import Session

import seedfarmer.errors
from seedfarmer.services._service_utils import boto3_client, boto3_resource, get_region

_logger: logging.Logger = logging.getLogger(__name__)


def get_role(
    role_name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> Optional[Dict[str, Any]]:
    _logger.debug("Getting Role: %s", role_name)

    iam_client = boto3_client("iam", session=session)
    try:
        return cast(Dict[str, Any], iam_client.get_role(RoleName=role_name))
    except iam_client.exceptions.NoSuchEntityException:
        return None


def get_role_arn(role_name: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> str:
    role = get_role(role_name=role_name, session=session)

    if not role:
        raise seedfarmer.errors.InvalidConfigurationError(f"Role {role_name} does not exist")

    return cast(str, role["Role"]["Arn"])


def create_check_iam_role(
    project_name: str,
    deployment_name: str,
    trust_policy: Dict[str, Any],
    role_name: str,
    permissions_boundary_arn: Optional[str],
    group_name: Optional[str] = None,
    module_name: Optional[str] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    role_prefix: Optional[str] = None,
) -> None:
    _logger.debug("Creating IAM Role with name: %s ", role_name)
    iam_client = boto3_client("iam", session=session)
    try:
        iam_client.get_role(RoleName=role_name)
    except iam_client.exceptions.NoSuchEntityException:
        args: Dict[str, Any] = {
            "RoleName": role_name,
            "Path": role_prefix if role_prefix else "/",
            "AssumeRolePolicyDocument": json.dumps(trust_policy),
            "Description": f"deployment-role for {role_name}",
            "Tags": [
                {"Key": "Region", "Value": get_region(session=session)},
                {"Key": "SeedFarmerProject", "Value": project_name},
                {"Key": "SeedFarmerDeployment", "Value": deployment_name},
            ],
        }
        if group_name and module_name:
            args["Tags"].append({"Key": "SeedFarmerModule", "Value": f"{group_name}-{module_name}"})
        if permissions_boundary_arn:
            args["PermissionsBoundary"] = permissions_boundary_arn
        iam_client.create_role(**args)


def attach_policy_to_role(
    role_name: str, policies: List[str], session: Optional[Union[Callable[[], Session], Session]] = None
) -> List[str]:
    iam_client = boto3_client("iam", session=session)
    try:
        response = iam_client.list_attached_role_policies(RoleName=role_name)
        attached_policies = [i["PolicyArn"] for i in response["AttachedPolicies"] if response["AttachedPolicies"]]
        _logger.debug("List of Attached Policie(s) are: %s ", attached_policies)
        to_be_attached = list(set(policies) - set(attached_policies))
        # if attached_polices is empty, to_bo_attached == policies

        if attached_policies:
            _logger.debug("To be attached policie(s) are %s ", to_be_attached)
        else:
            _logger.debug("First time deployment, attaching all policies: %s ", to_be_attached)

        for policy in to_be_attached:
            iam_client.attach_role_policy(RoleName=role_name, PolicyArn=policy)
            _logger.debug("Attached the policy: %s ", policy)
        return to_be_attached
    except Exception as e:
        raise e


def attach_inline_policy(
    role_name: str, policy_body: str, policy_name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    iam_client = boto3_client("iam", session=session)
    _logger.debug("Attaching the Inline policy %s to the IAM Role: %s", policy_name, role_name)
    try:
        iam_client.put_role_policy(RoleName=role_name, PolicyName=policy_name, PolicyDocument=policy_body)
    except Exception as e:
        _logger.debug("ailed to attach the Inline policy %s to the IAM Role: %s", policy_name, role_name)
        raise e


def detach_policy_from_role(
    role_name: str, policy_arn: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    _logger.debug("Detatching policy: %s  from the IAM Role: %s ", policy_arn, role_name)
    iam_client = boto3_client("iam", session=session)
    try:
        iam_client.detach_role_policy(RoleName=role_name, PolicyArn=policy_arn)
    except Exception as e:
        _logger.info("Could not delete policy %s", policy_arn)
        _logger.info("Caught exception: %s ", e)


def delete_role(role_name: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> None:
    iam_client = boto3_client("iam", session=session)
    try:
        iam_client.delete_role(RoleName=role_name)
    except iam_client.exceptions.NoSuchEntityException as e:
        _logger.info("Could not delete role %s", role_name)
        _logger.info("Caught exception: %s ", e)


def detach_inline_policy_from_role(
    role_name: str, policy_name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    iam_resource = boto3_resource("iam", session)
    try:
        iam_resource.RolePolicy(role_name, policy_name).delete()
    except Exception as e:
        raise e


def get_policy_info(policy_arn: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> Dict[str, Any]:
    iam_client = boto3_client("iam", session=session)
    try:
        return cast(Dict[str, Any], iam_client.get_policy(PolicyArn=policy_arn))
    except iam_client.exceptions.NoSuchEntityException as ne:
        _logger.info("Policy does not exist: %s ", policy_arn)
        raise ne
    except Exception as e:
        raise e



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from pydantic import PrivateAttr

import seedfarmer.errors
from seedfarmer.models._base import CamelModel, ValueFromRef
from seedfarmer.models.manifests._module_manifest import ModuleManifest


class ModulesManifest(CamelModel):
    """
    ModulesManifest
    This is a group in the deployment and defines the name of the group,
    the path of the group definiton, and the modules that comprise the group
    """

    name: str
    path: Optional[str] = None
    modules: List[ModuleManifest] = []
    concurrency: Optional[int] = None


class NetworkMapping(CamelModel):
    """
    NetworkMapping
    This class provides network metadata
    """

    vpc_id: Optional[Union[str, ValueFromRef]] = None
    private_subnet_ids: Optional[Union[List[str], ValueFromRef]] = None
    security_group_ids: Optional[Union[List[str], ValueFromRef]] = None


class RegionMapping(CamelModel):
    """
    RegionMapping
    This class provides metadata about the regions where Modules are deployed
    """

    region: str
    default: bool = False
    parameters_regional: Dict[str, Any] = {}
    network: Optional[NetworkMapping] = None
    codebuild_image: Optional[str] = None
    runtime_overrides: Optional[Dict[str, str]] = None
    npm_mirror: Optional[str] = None
    pypi_mirror: Optional[str] = None
    pypi_mirror_secret: Optional[str] = None
    npm_mirror_secret: Optional[str] = None
    seedkit_metadata: Optional[Dict[str, Any]] = None
    seedfarmer_artifact_bucket: Optional[str] = None
    role_prefix: Optional[str] = None
    policy_prefix: Optional[str] = None


class TargetAccountMapping(CamelModel):
    """
    TargetAccountMapping
    This class provides metadata about the accounts where Modules are deployed
    """

    alias: str
    account_id: Union[int, str, ValueFromRef]
    default: bool = False
    parameters_global: Dict[str, str] = {}
    region_mappings: List[RegionMapping] = []
    codebuild_image: Optional[str] = None
    runtime_overrides: Optional[Dict[str, str]] = None
    npm_mirror: Optional[str] = None
    npm_mirror_secret: Optional[str] = None
    pypi_mirror: Optional[str] = None
    pypi_mirror_secret: Optional[str] = None
    _default_region: Optional[RegionMapping] = PrivateAttr(default=None)
    _region_index: Dict[str, RegionMapping] = PrivateAttr(default_factory=dict)
    role_prefix: Optional[str] = None
    policy_prefix: Optional[str] = None

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        for r in self.region_mappings:
            if r.default:
                self._default_region = r
            self._region_index[r.region] = r

    def get_region_mapping(self, region: str) -> Optional[RegionMapping]:
        return self._region_index.get(region, None)

    @property
    def actual_account_id(self) -> str:
        if isinstance(self.account_id, str) or isinstance(self.account_id, int):
            return str(self.account_id)
        elif isinstance(self.account_id, ValueFromRef):
            if self.account_id.value_from and self.account_id.value_from.module_metadata is not None:
                raise seedfarmer.errors.InvalidManifestError(
                    "Loading value from Module Metadata is not supported in the Deployment Manifest"
                )
            elif self.account_id.value_from and self.account_id.value_from.env_variable:
                account_id = os.getenv(self.account_id.value_from.env_variable, None)
                if account_id is None:
                    raise seedfarmer.errors.InvalidManifestError(
                        (
                            "Unable to resolve AccountId from Environment Variable:"
                            f" {self.account_id.value_from.env_variable}"
                        )
                    )
                return account_id
            else:
                raise seedfarmer.errors.InvalidManifestError("Unsupported valueFrom type")
        else:
            raise seedfarmer.errors.InvalidManifestError("Unsupported accountId type")

    @property
    def default_region_mapping(self) -> Optional[RegionMapping]:
        return self._default_region


class NameGenerator(CamelModel):
    """
    NameGenerator
    This class decrbites how to dynamically generate the name of a DeploymentManifest
    """

    prefix: Union[str, ValueFromRef]
    suffix: Union[str, ValueFromRef]

    def _get_value(self, value: Union[str, ValueFromRef]) -> str:  # type: ignore[override]
        if isinstance(value, str):
            return value
        elif isinstance(value, ValueFromRef):
            if value.value_from and value.value_from.module_metadata is not None:
                raise seedfarmer.errors.InvalidManifestError(
                    "Loading value from Module Metadata is not supported on a NameGenerator"
                )
            elif value.value_from and value.value_from.env_variable:
                env_value = os.getenv(value.value_from.env_variable, None)
                if env_value is None:
                    raise seedfarmer.errors.InvalidManifestError(
                        (f"Unable to resolve value from Environment Variable: {value.value_from.env_variable}")
                    )
                return env_value
            else:
                raise seedfarmer.errors.InvalidManifestError("Unsupported valueFrom type")
        else:
            raise seedfarmer.errors.InvalidManifestError("Unsupported value type")

    def generate_name(self) -> str:
        prefix = self._get_value(self.prefix)
        suffix = self._get_value(self.suffix)

        return f"{prefix}{suffix}"


class DeploymentManifest(CamelModel):
    """
    DeploymentManifest
    This represents the top layer of the Deployment definiton.
    It includes things like the name of the deployment, the groups in the deployment
    and a policy that is applied to all build roles.
    """

    name: Optional[str] = None
    name_generator: Optional[NameGenerator] = None
    toolchain_region: str
    groups: List[ModulesManifest] = []
    description: Optional[str] = None
    target_account_mappings: List[TargetAccountMapping] = []
    force_dependency_redeploy: Optional[bool] = False
    archive_secret: Optional[str] = None
    _default_account: Optional[TargetAccountMapping] = PrivateAttr(default=None)
    _account_alias_index: Dict[str, TargetAccountMapping] = PrivateAttr(default_factory=dict)
    _account_id_index: Dict[str, TargetAccountMapping] = PrivateAttr(default_factory=dict)
    _accounts_regions: Optional[List[Dict[str, str]]] = PrivateAttr(default=None)
    _module_index: Dict[Tuple[str, str], ModuleManifest] = PrivateAttr(default_factory=dict)
    _partition: Optional[str] = PrivateAttr(default="aws")

    def __init__(self, **kwargs: Any) -> None:
        from seedfarmer.utils import batch_replace_env

        kwargs = batch_replace_env(payload=kwargs)
        super().__init__(**kwargs)

        if self.name is None and self.name_generator is None:
            raise seedfarmer.errors.InvalidManifestError("One of 'name' or 'name_generator' is required")

        if self.name is not None and self.name_generator is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'name' or 'name_generator' can be specified")

        # Generate a name and then reset the name_generator to None so that any SerDe done later on does not
        # generate a new name
        if self.name_generator is not None:
            self.name = self.name_generator.generate_name()
            self.name_generator = None

        for ta in self.target_account_mappings:
            if ta.default:
                self._default_account = ta
            self._account_alias_index[ta.alias] = ta
            self._account_id_index[ta.actual_account_id] = ta

    def get_target_account_mapping(
        self, account_alias: Optional[str] = None, account_id: Optional[str] = None
    ) -> Optional[TargetAccountMapping]:
        if account_alias is None and account_id is None:
            raise seedfarmer.errors.InvalidManifestError("One of 'account_alias' or 'account_id' is required")
        elif account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")
        elif account_alias:
            return self._account_alias_index.get(account_alias, None)
        elif account_id:
            return self._account_id_index.get(account_id, None)
        else:
            return None

    @property
    def default_target_account_mapping(self) -> Optional[TargetAccountMapping]:
        return self._default_account

    @property
    def target_accounts_regions(self) -> List[Dict[str, str]]:
        if self._accounts_regions is None:
            self._accounts_regions = []
            for target_account in self.target_account_mappings:
                for region in target_account.region_mappings:
                    role_prefix = region.role_prefix if region.role_prefix else target_account.role_prefix
                    policy_prefix = region.policy_prefix if region.policy_prefix else target_account.policy_prefix
                    account_region_args = {
                        "alias": target_account.alias,
                        "account_id": target_account.actual_account_id,
                        "region": region.region,
                        "network": region.network,
                        "parameters_regional": region.parameters_regional,
                        "codebuild_image": cast(str, region.codebuild_image),
                        "role_prefix": role_prefix,
                        "policy_prefix": policy_prefix,
                    }
                    self._accounts_regions.append(account_region_args)  # type: ignore
        return self._accounts_regions

    def get_parameter_value(
        self,
        parameter: str,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
        default: Optional[str] = None,
    ) -> Optional[Any]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None

        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                # Search the region_mappings for the region, if the parameter is in parameters_regional return it
                for region_mapping in target_account.region_mappings:
                    if (
                        region == region_mapping.region or (use_default_region and region_mapping.default)
                    ) and parameter in region_mapping.parameters_regional:
                        return region_mapping.parameters_regional[parameter]

                # If no region_mapping found for the region or no value for the parameter found in parameters_regional
                # return the value for parameter from parameters_global, default None
                return target_account.parameters_global.get(parameter, default)
        else:
            return default

    def get_region_codebuild_image(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[str]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None

        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                # Search the region_mappings for the region, if the codebuild_image is in region
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        image = (
                            region_mapping.codebuild_image
                            if region_mapping.codebuild_image is not None
                            else target_account.codebuild_image
                        )
                        if (
                            image is None
                            and region_mapping.seedkit_metadata
                            and region_mapping.seedkit_metadata.get("CodeBuildProjectBuildImage")
                        ):
                            image = region_mapping.seedkit_metadata["CodeBuildProjectBuildImage"]

                        return image
        else:
            return None

    def get_region_runtime_overrides(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[Dict[str, str]]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None

        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                # Search the region_mappings for the region, if the codebuild_image is in region
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        region_overrides = region_mapping.runtime_overrides if region_mapping.runtime_overrides else {}
                        account_overrides = target_account.runtime_overrides if target_account.runtime_overrides else {}
                        return {**account_overrides, **region_overrides}
        else:
            return None

    def get_region_npm_mirror(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[str]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None
        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                # Search the region_mappings for the region, if the npm_mirror is in region
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        npm_mirror = (
                            region_mapping.npm_mirror
                            if region_mapping.npm_mirror is not None
                            else target_account.npm_mirror
                        )
                        return npm_mirror
        else:
            return None

    def get_region_pypi_mirror(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[str]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None
        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                # Search the region_mappings for the region, if the pypi_mirror is in region
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        pypi_mirror = (
                            region_mapping.pypi_mirror
                            if region_mapping.pypi_mirror is not None
                            else target_account.pypi_mirror
                        )
                        return pypi_mirror
        else:
            return None

    def get_region_mirror_secret(
        self,
        *,
        mirror_type: Optional[str] = "pypi",
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[str]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        if mirror_type not in ["pypi", "npm"]:
            raise seedfarmer.errors.InvalidManifestError("Mirror type must be of type 'npm' or 'pypi'")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None
        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                # Search the region_mappings for the region, if the [pypi|npm]_mirror_secret is in region
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        if mirror_type == "pypi":
                            pypi_mirror_secret = (
                                region_mapping.pypi_mirror_secret
                                if region_mapping.pypi_mirror_secret is not None
                                else target_account.pypi_mirror_secret
                            )
                            return pypi_mirror_secret
                        elif mirror_type == "npm":
                            npm_mirror_secret = (
                                region_mapping.npm_mirror_secret
                                if region_mapping.npm_mirror_secret is not None
                                else target_account.npm_mirror_secret
                            )
                            return npm_mirror_secret
        else:
            return None

    def get_region_seedfarmer_bucket(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[str]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None
        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        sf_bucket = (
                            region_mapping.seedfarmer_artifact_bucket
                            if region_mapping.seedfarmer_artifact_bucket is not None
                            else None
                        )
                        return sf_bucket
        else:
            return None

    def get_region_seedfarmer_metadata(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Optional[Dict[str, str]]:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None
        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        return region_mapping.seedkit_metadata if region_mapping.seedkit_metadata else None
        else:
            return None

    def get_account_region_role_prefix(
        self,
        *,
        account_alias: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> str:
        if account_alias is not None and account_id is not None:
            raise seedfarmer.errors.InvalidManifestError("Only one of 'account_alias' and 'account_id' is allowed")

        use_default_account = account_alias is None and account_id is None
        use_default_region = region is None
        default_prefix = "/"
        for target_account in self.target_account_mappings:
            if (
                account_alias == target_account.alias
                or account_id == target_account.actual_account_id
                or (use_default_account and target_account.default)
            ):
                for region_mapping in target_account.region_mappings:
                    if region == region_mapping.region or (use_default_region and region_mapping.default):
                        role_prefix = (
                            region_mapping.role_prefix
                            if region_mapping.role_prefix is not None
                            else target_account.role_prefix
                        )
                        return role_prefix if role_prefix else default_prefix
        else:
            return default_prefix

    def get_permission_boundary_arn(self, target_account: str, target_region: str) -> Optional[str]:
        permissions_boundary_name = self.get_parameter_value(
            "permissionsBoundaryName",
            account_id=target_account,
            region=target_region,
        )
        return (
            f"arn:{self._partition}:iam::{target_account}:policy/{permissions_boundary_name}"
            if permissions_boundary_name is not None
            else None
        )

    def validate_and_set_module_defaults(self) -> None:
        for group in self.groups:
            for module in group.modules:
                self._module_index[(group.name, module.name)] = module
                module.target_account = (
                    self.default_target_account_mapping.alias
                    if self.default_target_account_mapping is not None and module.target_account is None
                    else module.target_account
                )

                target_account = self.get_target_account_mapping(account_alias=module.target_account)
                if target_account is None:
                    raise seedfarmer.errors.InvalidManifestError(
                        f"Invalid target_account ({module.target_account}) for "
                        f"Module {module.name} in Group {group.name}"
                    )
                module.set_target_account_id(target_account.actual_account_id)

                module.target_region = (
                    target_account.default_region_mapping.region
                    if target_account.default_region_mapping is not None and module.target_region is None
                    else module.target_region
                )

                if target_account.get_region_mapping(region=cast(str, module.target_region)) is None:
                    raise seedfarmer.errors.InvalidManifestError(
                        f"Invalid target_region ({module.target_region}) in target_account ({target_account.alias}) "
                        f"for Module {module.name} in Group {group.name}"
                    )

    def get_module(self, group: str, module: str) -> Optional[ModuleManifest]:
        return self._module_index.get((group, module), None)

    def populate_metadata(self, account_id: str, region: str, seedkit_dict: Dict[str, Any]) -> None:
        for target_account in self.target_account_mappings:
            for region_mapping in target_account.region_mappings:
                if target_account.actual_account_id == account_id and region_mapping.region == region:
                    region_mapping.seedkit_metadata = seedkit_dict
                    region_mapping.seedfarmer_artifact_bucket = (
                        seedkit_dict.get("SeedfarmerArtifactBucket")
                        if seedkit_dict.get("SeedfarmerArtifactBucket")
                        else None
                    )
                    break



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import sys
from typing import Any, Dict, Optional

import yaml

from seedfarmer import config
from seedfarmer.utils import create_output_dir

_logger: logging.Logger = logging.getLogger(__name__)

# FILENAME = "template.yaml"
# RESOURCES_FILENAME = os.path.join(CLI_ROOT, "resources", FILENAME)


def synth(
    *,
    deploy_id: str,
    synthesize: bool = False,
    **kwargs: Dict[str, Any],
) -> Optional[str]:
    out_dir = create_output_dir(f"seedkit-{deploy_id}")
    output_filename = os.path.join(out_dir, config.SEEDKIT_YAML_FILENAME)
    kwargs = {} if kwargs is None else kwargs

    _logger.debug("Reading %s", config.SEEDKIT_TEMPLATE_PATH)

    with open(config.SEEDKIT_TEMPLATE_PATH) as f:
        template = yaml.safe_load(f)

    if not synthesize:
        _logger.debug("Writing %s", output_filename)
        os.makedirs(os.path.dirname(output_filename), exist_ok=True)
        with open(output_filename, "w") as file:
            file.write(yaml.dump(template))
        return output_filename
    else:
        sys.stdout.write(yaml.dump(template))
        return None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import sys
from typing import List, Optional

import boto3
import click

import seedfarmer.messages as messages
import seedfarmer.mgmt.build_info as bi
import seedfarmer.mgmt.deploy_utils as du
import seedfarmer.mgmt.module_info as mi
from seedfarmer import DEBUG_LOGGING_FORMAT, commands, config, enable_debug
from seedfarmer.output_utils import (
    print_bolded,
    print_dependency_list,
    print_deployment_inventory,
    print_json,
    print_manifest_inventory,
)
from seedfarmer.services import get_sts_identity_info
from seedfarmer.services.session_manager import (
    ISessionManager,
    SessionManager,
    bind_session_mgr,
)
from seedfarmer.utils import load_dotenv_files

_logger: logging.Logger = logging.getLogger(__name__)


def _load_project() -> str:
    try:
        _logger.info("No --project provided, attempting load from seedfarmer.yaml")
        return config.PROJECT
    except FileNotFoundError:
        print_bolded("Unable to determine project to bootstrap, one of --project or a seedfarmer.yaml is required")
        raise click.ClickException("Failed to determine project identifier")


def _error_messaging(deployment: str, group: Optional[str] = None, module: Optional[str] = None) -> None:
    if group and module:
        print(f"No module data found for {deployment}-{group}-{module}")
        print_bolded(f"To see all deployed modules in {deployment}, run seedfarmer list modules -d {deployment}")
    else:
        print(f"No module data found for {deployment}")
    print_bolded("To see all deployments, run seedfarmer list deployments")


@click.group(name="list", help="List the relative data (module or deployment)")
def list() -> None:
    """List module data"""
    pass


@list.command(name="dependencies", help="List all dependencies of a module")
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_dependencies(
    deployment: str,
    group: str,
    module: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are fetching the dependencies for %s in %s of deployment %s", module, group, deployment)

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    SessionManager().get_or_create(project_name=project, profile=profile, region_name=region, qualifier=qualifier)
    dep_manifest = du.generate_deployed_manifest(deployment_name=deployment, skip_deploy_spec=True)

    if dep_manifest:
        module_depends_on_dict, module_dependencies_dict = du.generate_dependency_maps(manifest=dep_manifest)
        (
            print_dependency_list(
                header_message=f"Modules that {module} in {group} of {deployment} DEPENDS ON : ",
                modules=module_depends_on_dict[f"{group}-{module}"],
            )
            if module_depends_on_dict.get(f"{group}-{module}")
            else None
        )
        (
            print_dependency_list(
                header_message=f"Modules that ARE DEPENDENT ON {module} in {group} of {deployment} : ",
                modules=module_dependencies_dict[f"{group}-{module}"],
            )
            if module_dependencies_dict.get(f"{group}-{module}")
            else None
        )


@list.command(name="deployspec", help="List the stored deployspec of a module")
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_deployspec(
    deployment: str,
    group: str,
    module: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are getting deployspec for  %s in %s of deployment %s", module, group, deployment)

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    session_manager: ISessionManager = SessionManager().get_or_create(
        project_name=project, profile=profile, region_name=region, qualifier=qualifier
    )
    dep_manifest = du.generate_deployed_manifest(deployment_name=deployment, skip_deploy_spec=True)

    if dep_manifest is None:
        print_json({})
        return

    dep_manifest.validate_and_set_module_defaults()
    try:
        session = session_manager.get_deployment_session(
            account_id=dep_manifest.get_module(group=group, module=module).get_target_account_id(),  # type: ignore
            region_name=dep_manifest.get_module(group=group, module=module).target_region,  # type: ignore
        )
    except Exception:
        _error_messaging(deployment, group, module)
        return

    val = mi.get_deployspec(deployment=deployment, group=group, module=module, session=session)
    print_json(val)


@list.command(name="moduledata", help="Fetch the module metadata")
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--export-local-env/--no-export-local-env",
    default=False,
    help="Print the moduledata as env parameters for local development support INSTEAD of json (default is FALSE)",
    show_default=True,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_module_metadata(
    deployment: str,
    group: str,
    module: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    export_local_env: bool,
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are getting module data for  %s in %s of %s", module, group, deployment)

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    session_manager: ISessionManager = SessionManager().get_or_create(
        project_name=project, profile=profile, region_name=region, qualifier=qualifier
    )
    dep_manifest = du.generate_deployed_manifest(deployment_name=deployment, skip_deploy_spec=True)

    if dep_manifest is None:
        _error_messaging(deployment, group, module)
        return

    dep_manifest.validate_and_set_module_defaults()
    try:
        session = session_manager.get_deployment_session(
            account_id=dep_manifest.get_module(group=group, module=module).get_target_account_id(),  # type: ignore
            region_name=dep_manifest.get_module(group=group, module=module).target_region,  # type: ignore
        )
    except Exception:
        _error_messaging(deployment, group, module)
        return

    metadata_json = mi.get_module_metadata(deployment, group, module, session=session)
    if not export_local_env:
        sys.stdout.write(json.dumps(metadata_json))
    else:
        envs = commands.generate_export_env_params(metadata_json)
        if envs:
            for exp in envs:
                sys.stdout.write(exp)
                sys.stdout.write("\n")
        else:
            _error_messaging(deployment, group, module)


@list.command(name="allmoduledata", help="Fetch ALL module metadata in the deployment as a dict")
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_all_module_metadata(
    deployment: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are getting all module data for  %s", deployment)

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    session = SessionManager().get_or_create(
        project_name=project, profile=profile, region_name=region, qualifier=qualifier
    )
    dep_manifest = du.generate_deployed_manifest(deployment_name=deployment, skip_deploy_spec=True)

    if dep_manifest is None:
        _error_messaging(deployment)
        return
    dep_manifest.validate_and_set_module_defaults()
    try:
        all_metadata_json = {
            f"{group.name}-{module.name}": mi.get_module_metadata(
                deployment=deployment,
                group=group.name,
                module=module.name,
                session=(
                    session.get_deployment_session(
                        account_id=module.get_target_account_id(),  # type: ignore
                        region_name=module.target_region,  # type: ignore
                    )
                ),
            )
            for group in dep_manifest.groups
            for module in group.modules
        }
        sys.stdout.write(json.dumps(all_metadata_json))
    except Exception:
        _error_messaging(deployment)
        return


@list.command(name="modules", help="List the modules in a deployment")
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_modules(
    deployment: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are getting modules for %s", deployment)

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    SessionManager().get_or_create(project_name=project, profile=profile, region_name=region, qualifier=qualifier)

    dep_manifest = du.generate_deployed_manifest(deployment_name=deployment, skip_deploy_spec=True)
    if dep_manifest:
        print_manifest_inventory("Deployed Modules", dep_manifest, False, "green")


@list.command(name="deployments", help="List the deployments in this account")
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_deployments(
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    if project is None:
        project = _load_project()
    _logger.debug("Listing all deployments for Project %s", project)

    session_manager = SessionManager().get_or_create(
        project_name=project, profile=profile, region_name=region, qualifier=qualifier
    )
    deps = mi.get_all_deployments(session=session_manager.toolchain_session)
    if not deps or len(deps) == 0:
        account_id, _, _ = get_sts_identity_info(session=session_manager.toolchain_session)
        region = session_manager.toolchain_session.region_name
        _logger.info("No Deployments found for project %s in account %s and region %s", project, account_id, region)
        print_bolded(message=messages.no_deployment_found(), color="yellow")
    else:
        print_deployment_inventory(description="Deployment Names", dep=deps)


@list.command(
    name="buildparams",
    help="""Fetch the environment params of an executed build.
               This is to help with local development efforts.""",
)
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--build-id",
    type=str,
    help="The Build ID to fetch this info for",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--export-local-env/--no-export-local-env",
    default=False,
    help="Print the moduledata as env parameters for local development support INSTEAD of json (default is FALSE)",
    show_default=True,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def list_build_env_params(
    deployment: str,
    group: str,
    module: str,
    build_id: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    export_local_env: str,
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug(
        "We are getting build environment parameters for id %s of %s in %s of %s", build_id, module, group, deployment
    )

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    session_manager: ISessionManager = SessionManager().get_or_create(
        project_name=project, profile=profile, region_name=region, qualifier=qualifier
    )
    dep_manifest = du.generate_deployed_manifest(
        deployment_name=deployment, skip_deploy_spec=True, ignore_deployed=True
    )

    if dep_manifest is None:
        _error_messaging(deployment, group, module)
        return

    dep_manifest.validate_and_set_module_defaults()
    try:
        session: boto3.Session = session_manager.get_deployment_session(
            account_id=dep_manifest.get_module(group=group, module=module).get_target_account_id(),  # type: ignore
            region_name=dep_manifest.get_module(group=group, module=module).target_region,  # type: ignore
        )
    except Exception:
        _error_messaging(deployment, group, module)
        return

    metadata_json = bi.get_build_env_params(build_ids=[build_id], session=session)
    if not export_local_env:
        sys.stdout.write(json.dumps(metadata_json))
    else:
        envs = commands.generate_export_raw_env_params(metadata=metadata_json)
        if envs:
            for exp in envs:
                sys.stdout.write(exp)
                sys.stdout.write("\n")
        else:
            _error_messaging(deployment, group, module)


@list.command(
    name="schema",
    help="""Generate the schema that SeedFarmer uses for manifest objects.
             Either the deployment manifest or module manifest schema
             can be requested.
             This will return a formatted string of the schema that can
             be piped to a file. """,
)
@click.option(
    "--type",
    "-t",
    type=click.Choice(["deployment", "module"], case_sensitive=True),
    help="Either 'deployment' or 'module' can be used, default is `deployment`",
    required=False,
    default="deployment",
)
@bind_session_mgr
def list_manifest_schema(type: str) -> None:
    print(json.dumps(bi.get_manifest_schema(type=type), indent=2))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import subprocess
import uuid
from datetime import datetime, timezone
from typing import Dict, Optional

import seedfarmer.services._codebuild as codebuild

_logger: logging.Logger = logging.getLogger(__name__)


def run(
    local_deploy_path: str,
    env_vars: Dict[str, str] = {},
    codebuild_image: Optional[str] = None,
) -> codebuild.BuildInfo:
    # Write the environment variables to the file
    env_vars_path = os.path.join(local_deploy_path, "local.env")
    with open(env_vars_path, "w") as f:
        for key, value in env_vars.items():
            f.write(f"{key}={value}\n")

    ## Extract the zip to the local root so it is mounted by the container

    codebuild_image = (
        codebuild_image if codebuild_image else "public.ecr.aws/codebuild/amazonlinux2-x86_64-standard:5.0"
    )

    docker_command = [
        "docker",
        "run",
        "-v",
        "/var/run/docker.sock:/var/run/docker.sock",
        "-e",
        f"IMAGE_NAME={codebuild_image}",
        "-e",
        f"ARTIFACTS={local_deploy_path}/artifacts",
        "-e",
        f"SOURCE={local_deploy_path}/",
        "-e",
        f"BUILDSPEC={local_deploy_path}/buildspec/buildspec.yaml",
        "-v",
        f"{local_deploy_path}/:/LocalBuild/envFile/",
        "-e",
        "ENV_VAR_FILE=local.env",
        "-e",
        f"AWS_CONFIGURATION={os.environ.get('HOME')}/.aws",
        "-e",
        "AWS_EC2_METADATA_DISABLED=true",
        "-e",
        "MOUNT_SOURCE_DIRECTORY=TRUE",
        "-e",
        "INITIATOR=codebuild-user",
        "-e",
        f"COMPOSE_PROJECT_NAME=agent-{str(uuid.uuid4())[:8]}",
        "-e",
        f"REPORTS={local_deploy_path}/logs",
    ]

    for key, value in os.environ.items():
        if key.startswith("AWS_"):
            env_current = ["-e", f"{key}={value}"]
            docker_command.extend(env_current)

    docker_command.append("public.ecr.aws/codebuild/local-builds:latest")

    try:
        start_time = datetime.now(timezone.utc)
        _logger.debug("Running local Docker deployment: \n%s", " ".join(docker_command))

        process = subprocess.run(docker_command, check=True)

        end_time = datetime.now(timezone.utc)
        duration = (end_time - start_time).total_seconds()

        return codebuild.BuildInfo(
            build_id="local",
            status=codebuild.BuildStatus.succeeded if process.returncode == 0 else codebuild.BuildStatus.failed,
            start_time=start_time,
            end_time=end_time,
            duration_in_seconds=duration,
            current_phase=codebuild.BuildPhaseType.build,
            exported_env_vars=env_vars,
            phases=None,  # type: ignore[arg-type]
            logs=None,  # type: ignore[arg-type]
        )
    except KeyboardInterrupt:
        _logger.info("Interrupted by the user")
    except (subprocess.CalledProcessError, Exception) as e:
        _logger.error(e)

    end_time = datetime.now(timezone.utc)
    duration = (end_time - start_time).total_seconds()

    return codebuild.BuildInfo(
        build_id="local",
        status=codebuild.BuildStatus.failed,
        start_time=start_time,
        end_time=datetime.now(timezone.utc),
        duration_in_seconds=duration,
        current_phase=codebuild.BuildPhaseType.build,
        exported_env_vars=env_vars,
        phases=None,  # type: ignore[arg-type]
        logs=None,  # type: ignore[arg-type]
    )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import List, Optional

import click

import seedfarmer.mgmt.deploy_utils as du
import seedfarmer.mgmt.module_info as mi
from seedfarmer import DEBUG_LOGGING_FORMAT, config, enable_debug
from seedfarmer.output_utils import print_bolded
from seedfarmer.services.session_manager import (
    ISessionManager,
    SessionManager,
    bind_session_mgr,
)
from seedfarmer.utils import load_dotenv_files

_logger: logging.Logger = logging.getLogger(__name__)


def _load_project() -> str:
    try:
        _logger.info("No --project provided, attempting load from seedfarmer.yaml")
        return config.PROJECT
    except FileNotFoundError:
        print_bolded("Unable to determine project to bootstrap, one of --project or a seedfarmer.yaml is required")
        raise click.ClickException("Failed to determine project identifier")


def _error_messaging(deployment: str, group: Optional[str] = None, module: Optional[str] = None) -> None:
    if group and module:
        print(f"No module info found for {deployment}-{group}-{module}")
        print_bolded(f"To see all deployed modules in {deployment}, run seedfarmer list modules -d {deployment}")
    else:
        print(f"No module info found for {deployment}")
    print_bolded("To see all deployments, run seedfarmer list deployments")


@click.group(name="taint", help="Top Level command to support adding a taint to a deployed module")
@bind_session_mgr
def taint() -> None:
    """Taint module"""
    pass


@taint.command(
    name="module",
    help="""This command will mark a module as needing
        redeploy of a module on the next deployment.
        Do not use this unless you are sure of the ramifications!
        """,
)
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def taint_module(
    deployment: str,
    group: str,
    module: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    env_files: List[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are removing module data for %s of group %s in %s", module, group, deployment)

    if project is None:
        project = _load_project()

    load_dotenv_files(config.OPS_ROOT, env_files=env_files)

    session_manager: ISessionManager = SessionManager().get_or_create(
        project_name=project, profile=profile, region_name=region, qualifier=qualifier
    )

    try:
        dep_manifest = du.generate_deployed_manifest(deployment_name=deployment, skip_deploy_spec=True)
        dep_manifest.validate_and_set_module_defaults()  # type: ignore
        session = session_manager.get_deployment_session(
            account_id=dep_manifest.get_module(group=group, module=module).get_target_account_id(),  # type: ignore
            region_name=dep_manifest.get_module(group=group, module=module).target_region,  # type: ignore
        )
    except Exception:
        _error_messaging(deployment, group, module)
        return

    mi.remove_module_md5(
        deployment=deployment,
        group=group,
        module=module,
        type=mi.ModuleConst.BUNDLE,
        session=session,
    )
    _logger.debug("Module %s-%s-%s marked for redeploy", module, group, deployment)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import logging
from typing import List, Optional

import click

import seedfarmer
from seedfarmer import DEBUG_LOGGING_FORMAT, commands, config, enable_debug
from seedfarmer.cli_groups import bootstrap, bundle, init, list, metadata, projectpolicy, remove, seedkit, store, taint
from seedfarmer.output_utils import print_bolded
from seedfarmer.utils import load_dotenv_files

_logger: logging.Logger = logging.getLogger(__name__)


@click.group()
@click.version_option(seedfarmer.__version__)
def cli() -> None:
    """SeedFarmer CLI interface"""
    pass


@click.command(help="Get the version SeedFarmer")
def version() -> None:
    print(f"seed-farmer-{seedfarmer.__version__}")


@click.command(help="Apply manifests to a SeedFarmer managed deployment")
@click.argument(
    "spec",
    type=str,
    required=True,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--role-prefix",
    default=None,
    help="""An IAM path prefix of the seedfarmer toolchain and target roles.
    Use only if bootstrapped with this path""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
    type=bool,
)
@click.option(
    "--dry-run/--no-dry-run",
    default=False,
    help="Apply but do not execute....",
    show_default=True,
    type=bool,
)
@click.option(
    "--show-manifest/--no-show-manifest",
    default=False,
    help="Write out the generated deployment manifest to console",
    show_default=True,
    type=bool,
)
@click.option(
    "--enable-session-timeout/--disable-session-timeout",
    default=False,
    help="Enable boto3 Session timeouts. If enabled, boto3 Sessions will be reset on the timeout interval",
    show_default=True,
    type=bool,
)
@click.option(
    "--session-timeout-interval",
    default=900,
    help="If --enable-session-timeout, the interval, in seconds, to reset boto3 Sessions",
    show_default=True,
    type=int,
)
@click.option(
    "--update-seedkit/--no-update-seedkit",
    default=False,
    help="Force SeedFarmer to update the SeedKit when invoked",
    show_default=True,
    type=bool,
)
@click.option(
    "--update-project-policy/--no-update-project-policy",
    default=False,
    help="Force SeedFarmer to update the deployed Project Policy when invoked",
    show_default=True,
    type=bool,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
def apply(
    spec: str,
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    role_prefix: Optional[str],
    env_files: List[str],
    debug: bool,
    dry_run: bool,
    show_manifest: bool,
    enable_session_timeout: bool,
    session_timeout_interval: int,
    update_seedkit: bool,
    update_project_policy: bool,
    local: bool,
) -> None:
    """Apply manifests to a SeedFarmer managed deployment"""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)

    # Load environment variables from .env file if it exists
    load_dotenv_files(config.OPS_ROOT, env_files)

    _logger.info("Apply request with manifest %s", spec)
    if dry_run:
        print_bolded(" ***   This is a dry-run...NO ACTIONS WILL BE TAKEN  *** ", "white")

    commands.apply(
        deployment_manifest_path=spec,
        profile=profile,
        region_name=region,
        qualifier=qualifier,
        role_prefix=role_prefix,
        dryrun=dry_run,
        show_manifest=show_manifest,
        enable_session_timeout=enable_session_timeout,
        session_timeout_interval=session_timeout_interval,
        update_seedkit=update_seedkit,
        update_project_policy=update_project_policy,
        local=local,
    )


@click.command(help="Destroy a SeedFarmer managed deployment")
@click.argument(
    "deployment",
    type=str,
    required=True,
)
@click.option(
    "--dry-run/--no-dry-run",
    default=False,
    help="Apply but do not execute....",
    show_default=True,
)
@click.option(
    "--show-manifest/--no-show-manifest",
    default=False,
    help="Write out the generated deployment manifest",
    show_default=True,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
    Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--role-prefix",
    default=None,
    help="""An IAM path prefix of the seedfarmer toolchain and target roles.
    Use only if bootstrapped with this path""",
    required=False,
)
@click.option(
    "--env-file",
    "env_files",
    default=[".env"],
    help="""A relative path to the .env file to load environment variables from.
    Multple files can be passed in by repeating this flag, and the order will be
    preserved when overriding duplicate values.
    """,
    multiple=True,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--enable-session-timeout/--disable-session-timeout",
    default=False,
    help="Enable boto3 Session timeouts. If enabled, boto3 Sessions will be reset on the timeout interval",
    show_default=True,
    type=bool,
)
@click.option(
    "--session-timeout-interval",
    default=900,
    help="If --enable-session-timeout, the interval, in seconds, to reset boto3 Sessions",
    show_default=True,
    type=int,
)
@click.option(
    "--remove-seedkit/--no-remove-seedkit",
    default=False,
    help="""Delete the seedkit after destroy of deployment.
     NOTE: this will forcibly remove the seedkit for ALL deployments of this project.
     Use with CAUTION.  If you are unsure, do not use this flag.
    """,
    show_default=True,
    type=bool,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
def destroy(
    deployment: str,
    dry_run: bool,
    show_manifest: bool,
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    role_prefix: Optional[str],
    env_files: List[str],
    debug: bool,
    enable_session_timeout: bool,
    session_timeout_interval: int,
    remove_seedkit: bool,
    local: bool,
) -> None:
    """Destroy a SeedFarmer managed deployment"""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)

    # Load environment variables from .env file if it exists
    load_dotenv_files(config.OPS_ROOT, env_files)

    # MUST use seedfarmer.yaml so we can initialize configs
    project = config.PROJECT
    _logger.debug("Listing all deployments for Project %s", project)

    _logger.info("Destroy for Project %s, Deployment %s", project, deployment)
    if dry_run:
        print_bolded(" ***   This is a dry-run...NO ACTIONS WILL BE TAKEN  *** ", "white")

    commands.destroy(
        deployment_name=deployment,
        profile=profile,
        region_name=region,
        qualifier=qualifier,
        role_prefix=role_prefix,
        dryrun=dry_run,
        show_manifest=show_manifest,
        enable_session_timeout=enable_session_timeout,
        session_timeout_interval=session_timeout_interval,
        remove_seedkit=remove_seedkit,
        local=local,
    )


def main() -> int:
    cli.add_command(apply)
    cli.add_command(destroy)
    cli.add_command(store)
    cli.add_command(remove)
    cli.add_command(list)
    cli.add_command(init)
    cli.add_command(version)
    cli.add_command(bootstrap)
    cli.add_command(projectpolicy)
    cli.add_command(metadata)
    cli.add_command(bundle)
    cli.add_command(taint)
    cli.add_command(seedkit)
    cli()
    return 0
